/*
 * Copyright © 2014 f35f22fan@gmail.com
 *
 * Read the LICENSE file that comes with this project for license details.
*/

#ifndef ODS_BOOK_HPP_
#define ODS_BOOK_HPP_
#include "global.hxx"
#include "Content.hpp"
#include "decl.hxx"
#include "style/decl.hxx"
#include "Duration.hpp"
#include "err.hpp"
#include "i18n.hxx"
#include "ods.hh"
#include "tag.hxx"
#include <QDateTime>
#include <QString>
#include <QTemporaryDir>
#include <QVector>

class QXmlStreamWriter;

namespace ods	{

class ODS_API Book
{
public:
	Book();
	Book(const QString &path);
	virtual ~Book();
	
	void
	Add(ods::DrawFrame *df);
	
	ods::Cell*
	cell(const int row_no, const int col_no)
	{
		return cell(0, row_no, col_no);
	}
	
	ods::Cell*
	cell(const int sheet_no, const int row_no, const int col_no);
	
	ods::Content*
	content() { return content_; }
	
	ods::Style*
	CreateCellStyle() { return CreateStyle(ods::StyleFamilyId::Cell); }

	ods::Style*
	CreateColumnStyle(const ods::StylePlace place = ods::StylePlace::ContentFile)
	{
		return CreateStyle(ods::StyleFamilyId::Column, place);
	}
	
	ods::Style*
	CreateStyle(const ods::CurrencyInfo &info);

	ods::Style*
	CreateStyle(const ods::DateInfo &info);

	ods::Style*
	CreateStyle(const ods::DurationInfo &info);
	
	ods::style::Percent*
	CreatePercentStyle(const ods::StylePlace place
		= ods::StylePlace::ContentFile);
	
	ods::Style*
	CreateRowStyle() { return CreateStyle(ods::StyleFamilyId::Row); }
	
	ods::Sheet*
	CreateSheet(const QString &sheet_name);
	
	ods::Style*
	CreateStyle(const ods::StyleFamilyId,
		const ods::StylePlace = ods::StylePlace::StylesFile);
	
	ods::Style*
	CreateStyle(const ods::StyleFamilyId id, const ods::StylePlace place,
		ods::tag::func func);
	
	QVector<ods::DrawFrame*>&
	draw_frames() { return draw_frames_; }
	
	const QString&
	err() const { return err_; }
	
	void
	err_set(const QString &s) { err_ = s; }
	
	bool
	error() const { return err_.size() > 0; }
	
	bool
	extracted() { return extracted_file_paths_.size() > 0; }

	ods::style::Currency*
	GetCurrencyStyle(const ods::CurrencyInfo*);

	ods::style::Currency*
	GetCurrencyStyle(const QString &name);

	ods::style::Date*
	GetDateStyle(const ods::DateInfo *info);

	ods::style::Date*
	GetDateStyle(const QString &name);
	
	const QString&
	GetFrameId(const QString kFileName);
	
	QString*
	GetMediaDirPath();
	
	ods::style::Percent*
	GetPercentStyle(const QString &name, const qint8 decimal_places = -1);
	
	ods::style::Percent*
	GetPercentStyle(const qint8 decimal_places);

	ods::Style*
	GetStyle(const QString &name, const ods::StyleFamilyId id);

	ods::style::Duration*
	GetDurationStyle(const QString &name);

	ods::style::Duration*
	GetDurationStyle(const ods::DurationInfo*);
	
	void
	InitDefault();
	
	ods::Manifest*
	manifest() const { return manifest_; }
	
	ods::Meta*
	meta() { return meta_; }
	
	bool // utility to print a cell's value
	Print(const int sheet_no, const int row_no, const int col_no,
		const bool prepend_address = true);
	
	QString
	Save(const QFile &targetOdsFile);
	
	ods::Sheet*
	sheet(int index) { return content_->sheet(index); }
	
	ods::Sheet*
	sheet(const QString &name)
	{
		if (content_ == nullptr)
			mtl_line("(NULL): content_");
		return content_->sheet(name);
	}
	
	int
	sheet_count() { return content_->sheets().size(); }
	
	QVector<ods::Sheet*>&
	sheets() { return content_->sheets(); }
	
	ods::style::Manager*
	style_manager() { return style_manager_; }
	
	const QString&
	temp_dir_path() const { return temp_dir_path_; }
	
	ods::Value*
	value(const int row_no, const int col_no)
	{
		return value(0, row_no, col_no);
	}
	
	ods::Value*
	value(const int sheet_no, const int row_no, const int col_no);
	
	void
	WriteStartDocument(QXmlStreamWriter&);
	
private:
	NO_ASSIGN_COPY_MOVE(Book);

	ods::style::Currency*
	CreateCurrencyStyle(const ods::StylePlace place
		= ods::StylePlace::ContentFile);

	ods::style::Date*
	CreateDateStyle(const ods::StylePlace place
		= ods::StylePlace::ContentFile);

	ods::style::Duration*
	CreateDurationStyle(const ods::StylePlace place
		= ods::StylePlace::ContentFile);
	
	void
	InitTempDir();
	
	void
	PrepareDir(const QString &save_dir, QString &err);
	
	void
	Read(const QString &file_path);
	
	void
	SaveMimeTypeFile(const QString &dir_path);
	
	ods::Content			*content_ = nullptr;
	QVector<ods::style::Currency*> currency_styles_;
	QVector<ods::style::Date*> date_styles_;
	const bool				dev_mode_ = false;
	QVector<ods::DrawFrame*>	draw_frames_;
	QVector<ods::style::Duration*> duration_styles_;
	QString					err_;
	QStringList				extracted_file_paths_;
	ods::Manifest			*manifest_ = nullptr;
	QString					media_dir_path_;
	ods::Meta				*meta_ = nullptr;
	ods::Settings			*settings_ = nullptr;
	ods::style::Manager		*style_manager_ = nullptr;
	QVector<ods::Style*>	styles_;
	QVector<ods::style::Percent*> percent_styles_;
	QTemporaryDir			temp_dir_;
	QString					temp_dir_path_;
};

} // namespace ods
#endif
